/*
 * Copyright (c) 2003 David E. Bakken, his research students, and Washington State University.
 * Please see the file LICENSE.pdf for more details on terms of use.
 */

package GUI;

import java.util.Hashtable;
import java.io.File;
import javax.swing.filechooser.*;

/*********************************************************
* Creates a file chooser filter for selecting specific
* files in a file chooser dialog.
*
* @author Olav Haugan
*
* Date: 14.10.01
*********************************************************/
public class ExtensionFileFilter extends FileFilter
{

	private String description; //The file filter description
	private Hashtable filters = null; //The different filters
	
	/*********************************************************
	* Constructor. Initializes variables.
	* 
	*********************************************************/
	public ExtensionFileFilter()
	{
		filters = new Hashtable();
		setDescription("undefined");
	}
	
	/*********************************************************
	* Adds a file extension to the list of extensions.
	*********************************************************/
	public void addExtension( String extension )
	{
		filters.put(extension.toLowerCase(), this);
	}
	
	/*********************************************************
	* Returns the file filter description.
	*********************************************************/
	public String getDescription()
	{
		return description;
	}
	
	/*********************************************************
	* Sets the file filter description.
	*********************************************************/
	public void setDescription( String description )
	{
		this.description = description;
	}
	
	/*********************************************************
	* Called by the file chooser dialog to check if a file
	* should be shown or not. Function checks if the selected
	* file is included in the filter.
	*
	* @param File The file that are selected
	*********************************************************/
	public boolean accept(File f)
	{
		if( f != null)
		{
			if( f.isDirectory() )
			{
				return true;
			}
			String extension = getExtension( f );
			if( extension != null && filters.get( extension ) != null )
			{
				return true;
			}
		}
		return false;
	}
	
	/*********************************************************
	* Extracts a file's extension from its file name
	*
	* @param File The file that we want the extension from
	*********************************************************/
	protected String getExtension( File f )
	{
		if(f != null)
		{
			String filename = f.getName();
			int i = filename.lastIndexOf( '.' );
			if( i > 0 && i < filename.length() - 1 )
			{
				return filename.substring(i+1).toLowerCase();
			}
		}
		return null;
	}
}