/*
 * Copyright (c) 2003 David E. Bakken, his research students, and Washington State University.
 * Please see the file LICENSE.pdf for more details on terms of use.
 */

package ZEN_IDL.be;

import ZEN_IDL.ast.*;
import ZEN_IDL.be.symbolTableUtils.*;
import ZEN_IDL.be.helperUtils.*;
import util.*;

import java.io.*;
import java.util.*;

/*********************************************************
* Generates code for constant declarations
*
* @author Olav Haugan
*
* 14.10.01
*********************************************************/
public class MQCConstCodeGenerator
{
	private String packagePath; /* The package path for constructing file in correct direcory */
	private String packageStr; /* The package string */
	private String packageName; /* The name of this package */
	private String name; /* The name of the constant */
	private String value; /* The value of the constant */
	private String type; /* The type of the constant */	

	private boolean inInterface; /* true if inside an interface */	
	private PrintWriter pw; /* The outputstream to write to when inside an interface */	
	private MQCCodeGenHelper codeGenHelper; /* The code generator helper object */
	
	/*********************************************************
	* Constructor. Initializes variables.
	* 
	*********************************************************/
	public MQCConstCodeGenerator()
	{
		codeGenHelper = new MQCCodeGenHelper(null);		
		packagePath = null;
		packageStr = null;
		packageName = null;
		name = null;
		value = null;
		type = null;
		inInterface = false;
		pw = null;
	}
	
	/*********************************************************
	* Sets the package path variable
	* 
	* @param String The package path
	*
	*********************************************************/
	public void setPackagePath( String packagePath )
	{
		this.packagePath = packagePath;
	}

	/*********************************************************
	* Sets the package string variable
	* 
	* @param String The package string
	*
	*********************************************************/
	public void setPackageStr( String packageStr )
	{
		this.packageStr = packageStr;
	}

	/*********************************************************
	* Sets the package name
	* 
	* @param String The package name
	*
	*********************************************************/
	public void setPackageName( String packageName )
	{
		this.packageName = packageName;
	}

	/*********************************************************
	* Sets the name of the constant
	* 
	* @param String The name of the constant
	*
	*********************************************************/
	public void setName( String name )
	{
		this.name = name;
	}
	
	/*********************************************************
	* Sets the value of the constant
	* 
	* @param String The value of the constant
	*
	*********************************************************/
	public void setValue( String value )
	{
		this.value = value;
	}

	/*********************************************************
	* Sets the type of the constant
	* 
	* @param String The interface name
	*
	*********************************************************/
	public void setType( String type )
	{
		this.type = type;
	}

	/*********************************************************
	* Sets th inInterface variable to true or false
	* 
	* @param boolean true if inside interface, false otherwise
	*
	*********************************************************/
	public void setInInterface( boolean inInterface )
	{
		this.inInterface = inInterface;
	}
	
	/*********************************************************
	* Sets the outputstream to write to when inside an
	* interface.
	*
	* @param PrintWriter The outputstream
	*
	*********************************************************/
	public void setPW( PrintWriter pw )
	{
		this.pw = pw;
	}
	
	/*********************************************************
	* Generates the constant code
	*********************************************************/
	public void generateCode()
	{
		if(inInterface)
		{
			genInsideInterfaceCode();
		}
		else
		{
			genCode();
		}
	}

	/*********************************************************
	* Generates the constant code inside an interface
	*********************************************************/
	public void genInsideInterfaceCode()
	{
		pw.println("\tfinal public static " + codeGenHelper.mapIDL2Java(type) + " " + name + " = (" + codeGenHelper.mapIDL2Java(type) + ") " + value + ";");
	}

	/*********************************************************
	* Generates the constant code outside an interface
	*********************************************************/
	public void genCode()
	{
		PrintWriter pw = codeGenHelper.makeFile(packagePath, name );

		if(packagePath.length() != 0)
			pw.println(packageStr);
		pw.println();

		//todo: big switch to select what classes to inherit from
		pw.println("public interface " + name);
		pw.println("{");
		pw.println("\tfinal public static " + codeGenHelper.mapIDL2Java(type) + " value = (" + codeGenHelper.mapIDL2Java(type) + ") " + value + ";");
		pw.println("};");
		pw.println("// End of file.");
		pw.flush();
		pw.close();
	}
	
}

