/*
 * Copyright (c) 2003 David E. Bakken, his research students, and Washington State University.
 * Please see the file LICENSE.pdf for more details on terms of use.
 */

package ZEN_IDL.be;

import ZEN_IDL.ast.*;
import java.io.*;
import java.util.*;

//Note: template_type_spec and scoped_name are not currently handled.
public class XExceptionVisitor extends AbstractVisitor
{
    public Object visit( ASTbase_type_spec node, Object data )
    {
        VisitorState currentState = (VisitorState)data;
        Token firstToken = node.getFirstToken();
        Token lastToken = node.getLastToken();
        String type_spec = firstToken.toString();
        while( firstToken != lastToken )
        {
            firstToken = firstToken.next;
            type_spec += " " + firstToken.toString();
        }
				currentState.addStructTypeSpec( type_spec );
        return data;
    }
    public Object visit( ASTscoped_name node, Object data )
    {
        VisitorState currentState = (VisitorState)data;
	String type_spec = node.jjtGetChild(0).getFirstToken().toString();
	currentState.addStructTypeSpec( type_spec );
	return data;
    }
    public Object visit( ASTsimple_declarator node, Object data )
    {
        VisitorState currentState = (VisitorState)data;
        StructDeclarator simple_declarator = new StructDeclarator( node.getFirstToken().toString() );
        currentState.addStructDeclarator( simple_declarator );
				String type_spec = node.jjtGetParent().getFirstToken().toString();
        return data;
    }
    public Object visit( ASTarray_declarator node, Object data )
    {
        VisitorState currentState = (VisitorState)data;
        currentState.initializeArrayBracketsAndSize();
        data = node.childrenAccept( this, data );
        currentState = (VisitorState)data;
        String name = node.jjtGetChild(0).getFirstToken().toString();
        String arrayBracketsAndSize = currentState.getArrayBracketsAndSize();
        int dimensions = node.jjtGetNumChildren() - 1;
        StructDeclarator array_declarator = new StructDeclarator( name, arrayBracketsAndSize, dimensions );
        currentState.addStructDeclarator( array_declarator );
        return data;
    }   
    public Object visit( ASTfixed_array_size node, Object data )
    {
        VisitorState currentState = (VisitorState)data;
        String bracketsAndSize = "[" + node.jjtGetChild(0).getFirstToken().toString() + "]";
        currentState.addBracketAndSize( bracketsAndSize );
        return data;
    }
}

