// client.cpp,v 1.5 2002/01/29 20:21:07 okellogg Exp

#include "TimingC.h"
#include "ace/Get_Opt.h"
#include <time.h>
#include <math.h>
#include "CppTimer.h"

// Headers (too lazy to put in their own file)
double findStdev(int nonGC, int iCnt);
void   printArray(int a[], int alen, int offset);

ACE_RCSID(foo, client, "client.cpp,v 1.5 2002/01/29 20:21:07 okellogg Exp")

const char *ior = "file://timing.ior";
long maxIterations = 1000;  // message call count value

#define REPEAT_CNT 3
//  int      REPEAT_CNT = 3;    // number of times to repeat the main timing loop
  long long     dt_cnt     = 0;      // count of dt[] elements
  long long     dt_sum     = 0;      // E[x]  of dt histogram (unnormalized by cnt)
  long long     dt_sum2    = 0;      // E[x2] of dt histogram (unnormalized by cnt)
  float    dt_avg     = 0;      // avg time based upon dt
  double   stdev      = 0.0;
  double   dt_stdev   = 0.0;    // standard deviation
  double   dt_stdev2  = 0.0;    // standard deviation squared
  double   STDEV_ERR  = 99.9;   // value to set stdev is an error occurs
#define    DT_SIZE      500     // 500 for linux w/TR, 5000 for SaJe w/TR, 1000 for TINI w/TR
//int      DT_SIZE    = 500;    // 500 for linux w/TR, 5000 for SaJe w/TR, 1000 for TINI w/TR
#define    DTW_SIZE     300     // 300 for linux w/TR, 1000 for SaJe w/TR, 100 for TINI w/TR
//int      DTW_SIZE   = 300;    // 300 for linux w/TR, 1000 for SaJe w/TR, 100 for TINI w/TR
  int      dt[DT_SIZE];         // stores either individual event times or bins of event counts
  int      dtw[DTW_SIZE];       // stores either individual event times or bins of event counts
  int      dt_offset  = 0;       // dt[] offset, ie w/ binning dt[i] correspond to time i+dt_offset
  int      dt_binEvents;         // false -> ind. event times, true -> event cnt bins in dt[]
  float    nSigma     = 3.5;     // default width of the peak (in stdev)
  int      MAX_PEAK_WIDTH = 75;  // the timing peak should be found within the first MAX_PEAK_WIDTH bins
  double   MAX_OK_STDEV   = 4.0; // maximum value of an ok/good peak stdev
  int      ticksPerMilliSecond;


int
parse_args (int argc, char *argv[])
{
  ACE_Get_Opt get_opts (argc, argv, "c:k:");
  int c;

  while ((c = get_opts ()) != -1)
    switch (c)
      {
      case 'c':
        maxIterations = atoi( get_opts.opt_arg ());
        break;

      case 'k':
        ior = get_opts.opt_arg ();
        break;

      case '?':
      default:
        printf("%s\n", get_opts.opt_arg());
        ACE_ERROR_RETURN ((LM_ERROR,
                           "usage:  %s "
                           "-c <count> "
                           "-k <ior> "
                           "\n",
                           argv [0]),
                          -1);
      }
  // Indicates sucessful parsing of the command line
  return 0;
}

int
main (int argc, char *argv[])
{
  long     r_cnt[REPEAT_CNT];   // r_cnt[i]   == i_th dt_cnt
  float    r_avg[REPEAT_CNT];   // r_avg[i]   == i_th dt_avg
  double   r_stdev[REPEAT_CNT]; // r_stdev[i] == i_th stdev
  int      dt_b[1];
  int      timeoutSec = 600;     // timeout, in seconds, for the main timing loop
  char     info[256];

  unsigned long long     dt_time,  dt_time0;

  int      i;
  long     dt_delta;
  long     dtw_time, dtw_time0;
  long     dtw_delta;

  ACE_TRY_NEW_ENV
    {
      CORBA::ORB_var orb =
        CORBA::ORB_init (argc, argv, "" ACE_ENV_ARG_PARAMETER);
      ACE_TRY_CHECK;

      if (parse_args (argc, argv) != 0)
        return 1;

      CORBA::Object_var tmp =
        orb->string_to_object(ior ACE_ENV_ARG_PARAMETER);
      ACE_TRY_CHECK;

      timing::foo_var timing =
        timing::foo::_narrow(tmp.in () ACE_ENV_ARG_PARAMETER);
      ACE_TRY_CHECK;

      if (CORBA::is_nil (timing.in ()))
        {
          ACE_ERROR_RETURN ((LM_DEBUG,
                             "Nil timing::foo reference <%s>\n",
                             ior),
                            1);
        }

      if (maxIterations > DT_SIZE) {
        dt_binEvents = 1;     // dt[i] == # of events that took i ms to complete
      } else {
        dt_binEvents = 0;     // dt[i] == delta time of event i
      }

//===================
      timer_init(1000000);
      ticksPerMilliSecond = 1000;

      char *configID = "cfg:p0:";

      long long ssStartTime = 0;  // steady state start time
      long long ssTotalTime = 0;  // steady state total time
      long long startTime   = 0;  // start of "real" timing loop time
      long long totalTime   = 0;  // total time for the "real" timing loop
      CORBA::Long the_long;

      //
      // STEADY STATE TIMING LOOP
      //--------------------------
      //
      printf("---Steady State Begin---\n");

      // init variable needed to compute the stdev of time deltas of each call
      for (i = 0;  i < DT_SIZE;  i++) {
          dt[i] = 0;
      }
      for (i = 0;  i < DTW_SIZE;  i++) {
          dtw[i] = 0;
      }
      ssStartTime = timer_currentTime();
      dt_time0 = timer_currentTime();

      int    dt_min = 999999999;
      int    ss_min;
      int    ss_max;
      int    ss_upperLimit = 999999999;
      int    ss_nLoop = 3;
      int    ss_jcnt  = 150;//15000;
      for (int j = 0;  j < ss_nLoop;  j++) {
          dt_sum  = 0;
          dt_sum2 = 0;
          dt_cnt  = 0;
          ss_min  = 999999999;
          ss_max  = 0;
          dt_time0 = timer_currentTime();
          for (i = 0;  i < ss_jcnt;  i++) {
              the_long = timing->bar (i ACE_ENV_SINGLE_ARG_PARAMETER);
              dt_time  = timer_currentTime();
              dt_delta = dt_time - dt_time0;
              if (dt_delta < ss_min) ss_min = dt_delta;
              if (dt_delta > ss_max) ss_max = dt_delta;
              if (dt_delta < ss_upperLimit) {
                  dt_sum  += dt_delta;
                  dt_sum2 += (long) dt_delta * (long) dt_delta;
                  dt_cnt++;
              }
              dt_time0 = dt_time;
          }
          if (ss_min < dt_min) {
              dt_min = ss_min;
          }
          // semi-fragile code, probably should check that dt_cnt > 1.  -ADM
          dt_avg    = (float) dt_sum / dt_cnt;
          dt_stdev2 = (double)(dt_sum2 - (dt_sum * dt_sum)/dt_cnt) / (dt_cnt-1);
          if (dt_stdev2 >= 0.0 ) {
              dt_stdev = sqrt(dt_stdev2);
          } else {
              printf("Oops! Likely integer overflow -- dt_stdev2=  %.6lf  < 0.0 -- Setting stdev to %.4lf!\n", dt_stdev2, STDEV_ERR);
              dt_stdev = STDEV_ERR;
          }
          ss_upperLimit = (int) (dt_avg + nSigma * dt_stdev);
          printf(" - %d:   avg: %.4lf  stdev: %.6lf  min/max: %d-%d  upL: %d\n",
                  (j+1)*ss_jcnt, dt_avg, dt_stdev, ss_min, ss_max, ss_upperLimit);
      }
      float  dt_ssAvg   = (float) (dt_avg / ticksPerMilliSecond);
      double dt_ssStdev = dt_stdev / ticksPerMilliSecond;
      long   dt_ssCnt   = dt_cnt;

      ssTotalTime = (timer_currentTime() - ssStartTime);
      float ssAvgTime = (float) (ssTotalTime / ticksPerMilliSecond) / (ss_nLoop * ss_jcnt);

      printf("---Steady State Info---\n");
      printf("Overall:  %.4f    n/a     %d  %.4f s\n", ssAvgTime, (ss_nLoop * ss_jcnt), (float) (ssTotalTime/(1000*ticksPerMilliSecond)));
      printf("  ->end:  %.4lf  %.6lf  %ld\n", dt_ssAvg, dt_ssStdev, dt_ssCnt);
      printf("---Steady State End---\n");

      int dd = (int) (5.0 * (dt_ssStdev * ticksPerMilliSecond)) + 10;
      if (dd > (int) (.2 * DT_SIZE)) {
          dd = (int) (.2 * DT_SIZE);
      }
      dt_offset = dt_min - dd;
      if (dt_offset < 0) {
          dt_offset = 0;
      }

      //
      // MAIN TIMING LOOP
      //------------------
      //
      int  iterations = 0;
      unsigned long long stop_time;
      for (int r = 0;  r < REPEAT_CNT;  r++) {
          i = -1;
          dt_delta  = -1;
          dtw_delta = -1;
          try {
              // init variable needed to compute the stdev of time deltas of each call
              for (i = 0;  i < DT_SIZE;  i++) {
                  dt[i] = 0;
              }
              for (i = 0;  i < DTW_SIZE;  i++) {
                  dtw[i] = 0;
              }

              startTime = timer_currentTime();
              dt_time0  = startTime;
              dtw_time0 = dt_time0 / 1000;
              stop_time = dt_time0 + timeoutSec * 1000 * ticksPerMilliSecond;

              int d;
              iterations = maxIterations;

              for (i = 0; i < iterations; i++) {
                  the_long = timing->bar (i ACE_ENV_SINGLE_ARG_PARAMETER);
                  dt_time   = timer_currentTime();
                  dt_delta  = dt_time - dt_time0;
                  dt_time0  = dt_time;
                  dtw_time  = dt_time / 1000;
                  dtw_delta = dtw_time - dtw_time0;
                  dtw_time0 = dtw_time;
                  if (dt_binEvents) {
                      d = dt_delta - dt_offset;
                      if (d < DT_SIZE) {
                          dt[d]++;
                      } else {
                          dt[DT_SIZE-1]++;
                      }
                  } else {
                      // store raw (unbinned) times into dt[]
                      dt[i]  = dt_delta;
                  }
                  // wide events are always binned
                  if (dtw_delta < DTW_SIZE) {
                      dtw[dtw_delta]++;
                  } else {
                      dtw[DTW_SIZE-1]++;
                  }
                  if (dt_time > stop_time) {
                      iterations = i;
                      break;
                  }
              }
              totalTime = (timer_currentTime() - startTime);
              float avgTime = (float) (totalTime / ticksPerMilliSecond) / iterations;
              float ssDelta = ssAvgTime / avgTime;

              dt_stdev = findStdev(false, iterations);  //warning computes dt_cnt, dt_avg, dt_sum, dt_sum2
              dt_avg   /= ticksPerMilliSecond;
              dt_stdev /= ticksPerMilliSecond;

              sprintf(info, "@@@   Loop%d:  %d  %.4lf  %.6lf  %d  @@  r%.4lf  @@  %s\n",
                              r, (int)(totalTime/(1000*ticksPerMilliSecond)), avgTime, dt_stdev, iterations, ssDelta, configID);

              dt_stdev  = findStdev(true, iterations);  //warning computes dt_cnt, dt_avg, dt_sum, dt_sum2
              dt_avg   /= ticksPerMilliSecond;
              dt_stdev /= ticksPerMilliSecond;
              float dt_ssDelta = dt_ssAvg / dt_avg;

              printf(info);   // print non-GC info

              printf("@@@ gcLoop%d:  %d  %.4lf  %.6lf  %lld  @@  r%.4lf\n", 
                      r, (int)(totalTime/(1000*ticksPerMilliSecond)), dt_avg, dt_stdev, dt_cnt, dt_ssDelta);
              printf("  @@  %s\n", + configID);
          }
          catch (...) {
              printf("An exception occured--likely an array bounds exception.\n");
              printf("i: %d dt_offset/dt/dtw: %d  %ld  %ld\n", i, dt_offset, dt_delta, dtw_delta);
              dt_cnt   = -1;
              dt_avg   = 999999;
              dt_stdev = STDEV_ERR;
          }

          r_cnt[r]   = dt_cnt;
          r_avg[r]   = dt_avg;
          r_stdev[r] = dt_stdev;
      }

      printf("---Statistical Info---");
      printf("Loop s:  %.4lf  %.6lf  %ld\n", ssAvgTime, dt_ssStdev, dt_ssCnt);

      dt_cnt   = r_cnt[0];
      dt_avg   = r_avg[0];
      dt_stdev = r_stdev[0];
      for (int r = 0;  r < REPEAT_CNT;  r++) {
          printf("Loop %d:  %.4lf  %.6lf  %ld\n", r, r_avg[r], r_stdev[r], r_cnt[r]);
          if (r_avg[r] < dt_avg) {
              dt_cnt   = r_cnt[r];
              dt_avg   = r_avg[r];
              dt_stdev = r_stdev[r];
          }
      }
      printf("@@@ Summary: %.4lf  %.6lf  %lld  @@- %s\n", dt_avg, dt_stdev, dt_cnt, configID);


//===================

      orb->destroy (ACE_ENV_SINGLE_ARG_PARAMETER);
      ACE_TRY_CHECK;
    }
  ACE_CATCHANY
    {
      ACE_PRINT_EXCEPTION (ACE_ANY_EXCEPTION,
                           "Exception caught:");
      return 1;
    }
  ACE_ENDTRY;

  return 0;
}


    /**
     * findStdev -- find the standard deviation of a set of point.
     *
     * @param nonGC   if true, computate the stdev of only the
     *                non-Garbage Collected, if false, then use all events
     * @param iCnt    number of non-binned events in dt[] (see dt_binEvents)
     * @return double the raw standard deviation (ie, non-scaled value).
     */
    double findStdev(int nonGC, int iCnt) {
        //WARNING: global variables dt_cnt, dt_sum, dt_sum2 are all modified
        //within this routine (yes--an ugly hack...)

        int    i, iStart, iPeakStart, iStop;
        int    dt_max;
        int    *dt_b;  // bins
        double stdev2, stdev;

        if (dt_binEvents) {
            // bins already computed, just alias dt[]
            dt_b   = dt;
        } else {
            // need to compute number of bins needed
            dt_offset = 3600000;     // an hour (in ms)
            dt_max    =       0;
            for (i = 0;  i < iCnt;  i++) {
                if (dt[i] < dt_offset) dt_offset = dt[i];
                if (dt[i] > dt_max)    dt_max    = dt[i];
            }
            dt_offset--;
            dt_max++;
            if (dt_max > dt_offset + 2*DT_SIZE ) {
                dt_max = dt_offset + 2*DT_SIZE;
            }
            // create the bins
            dt_b = new int[dt_max - dt_offset + 1];
            // stuff the bins
            for (i = 0;  i < iCnt;  i++) {
                if (dt[i] - dt_offset < 2*DT_SIZE) {
                    dt_b[dt[i] - dt_offset]++;
                } else {
                    dt_b[2*DT_SIZE]++;
                }
            }
        }

        // Initially, use the first MAX_PEAK_WIDTH non-zero channels to compute
        // the average and stdev of the timing peak
        iStart = 0;
        while (dt_b[iStart] == 0  &&  iStart < DT_SIZE-1) {
          iStart++;
        }
        iPeakStart = iStart+1;
        while (dt_b[iPeakStart] < 2  &&  iPeakStart < DT_SIZE-1) {
          iPeakStart++;
        }

        iStop = DT_SIZE-1;
        if (nonGC) {
            // the last bin contains overflow values--discard it
            iStop--;
        }
        if (iStop > iPeakStart + MAX_PEAK_WIDTH) {
            iStop = iPeakStart + MAX_PEAK_WIDTH;
        }
        if (nonGC) {
            printf("iStart/iPeakStart/iStop: %d  %d  %d\n", (iStart+dt_offset), (iPeakStart+dt_offset), (iStop+dt_offset));
        }
        int iStop0;
        long dtb, dti;
        double x, x2;

        do {
            iStop0 = iStop;
            // sum up the dt[] bins
            dt_cnt  = 0;
            dt_sum  = 0;
            dt_sum2 = 0;
            for (i = iStart;  i < iStop;  i++) {
                dt_cnt  += dt_b[i];
                dtb = dt_b[i];
                dti = dt_offset + i;
                dt_sum  += dtb * dti;
                dt_sum2 += dtb * dti * dti;
                // check for overflowed values (they will cause dt_sum2 to go negative)
                if (dt_sum2 < 0) {
                    printf("Oops! Likely integer overflow problem.");
                    printf("dt_sum2 < 0: %lld at i=%d\n", dt_sum2, (dt_offset + i));
                }
            }

            // compute the average and its standard dev.
            dt_avg  = (float)   dt_sum  / dt_cnt;
            x  = (double) dt_sum;
            x2 = (double) dt_sum2;
            stdev2  = (x2 - (x * x)/dt_cnt) / (dt_cnt-1);
            if (stdev2 >= 0) {
                stdev = sqrt(stdev2);
            } else {
                printf("Oops! Likely integer overflow -- stdev2= %.6lf  < 0.0 -- Setting stdev to %.6lf!\n", stdev2, STDEV_ERR);
                stdev = STDEV_ERR;
            }
            if (nonGC) {
                iStop = (int) (dt_avg + (nSigma * stdev) + 0.5) - dt_offset;
                if (iStop >= DT_SIZE) {
                    iStop = DT_SIZE - 1;
                }
                if ((iStop >= iStop0) && (stdev > MAX_OK_STDEV)) {
                    // Compute an alternate upper limit for the first (nonGC)
                    // peak. The value (dt_avg - iStart) is *assumed* to be the
                    // bottom (lower-time) tail of the nonGC peak. This means
                    // that (dt_avg - iStart) should be another approximation
                    // to the value of 3.5 * nSigma (of the nonGC peak). If this
                    // value gives a smaller iStop use it.
                    int iStop2 = (int) ((dt_avg - dt_offset) + ((dt_avg - dt_offset) - iPeakStart));
                    printf("avg,iStop,iStop2: %.4lf  %d  %d\n", dt_avg, (dt_offset+iStop), (dt_offset+iStop2));
                    if (iStop2 < iStop) {
                        iStop = iStop2;
                    }
                }
                if (iStop < iStart + 1) {
                    iStop = iStart + 1;
                }
                printf("x-y,avg,stdev,iStop: %d-%d  %.6lf  %.6lf  %d\n", 
                        (dt_offset+iStart), (dt_offset+iStop0), dt_avg, stdev, (dt_offset+iStop));
            }
        } while (iStop < iStop0);

        // Pretty-Print the data
        if (nonGC) {
            printf("--- %d  %d ---non-GC---\n", (dt_offset+iStart), (dt_offset+(iStop0-1)));
            printArray(dt_b, DT_SIZE, dt_offset);
            printf("---Wide bins---\n");
            printArray(dtw, DTW_SIZE, 0);
        }

        return stdev;
    }

    /**
    * printArray -- a pretty printer for the dt/dtw arrays
    *
    * @param a     an array to print
    * @param offset  index offset of the array
    */
    void printArray(int a[], int alen, int offset)
    {
        int     lastZero = 0;
        int     inZeros  = 1;
        printf(" %d  %d\n",  offset, a[0]);
        for (int i = 1;  i < alen;  i++) {
            if (a[i] > 0) {
                if (inZeros) {
                    inZeros = 0;
                    if (i-1 > lastZero) {
                        printf("%d  %d\n", (offset+i-1), a[i-1]);
                    }
                }
                printf("%d  %d\n", (offset+i), a[i]);
            } else {
                if ( inZeros ) {
                    // do nothing
                } else {
                    inZeros  = 1;
                    lastZero = i;
                    printf("%d  %d\n", (offset+i), a[i]);
                }
            }
        }
    }
