/*
 * Copyright (c) 2003 David E. Bakken, his research students, and Washington State University.
 * Please see the file LICENSE.pdf for more details on terms of use.
 */

//include(`Config_macros.m4')

import mqc.*;
import mqc.holders.*;
import mqc.Config;

__SECURITY_ORBKEY__(
import mqc.security.InvalidKeyException;
)
__SAJE__(
import com.ajile.lang.Math;
)

public class Client
{
    static final int   REPEAT_CNT = __MAIN_LOOP_CNT__;       // number of times to repeat the main timing loop
    __DEBUG_TIMING__(
    static       long  dt_cnt     = 0;    // count of dt[] elements
    static       long  dt_sum     = 0;    // E[x]  of dt histogram (unnormalized by cnt)
    static       long  dt_sum2    = 0;    // E[x2] of dt histogram (unnormalized by cnt)
    static       float dt_avg     = 0;    // avg time based upon dt 
    static      double dt_stdev   = 0.0;  // standard deviation
    static      double dt_stdev2  = 0.0;  // standard deviation squared
    static  double STDEV_ERR      = 99.9; // value to set stdev is an error occurs
    static      long[] r_cnt      = new long[REPEAT_CNT];    // r_cnt[i]   == i_th dt_cnt
    static     float[] r_avg      = new float[REPEAT_CNT];   // r_avg[i]   == i_th dt_avg
    static    double[] r_stdev    = new double[REPEAT_CNT];  // r_stdev[i] == i_th stdev
    static final int   DT_SIZE    = 500;    // 500 for linux w/TR, 5000 for SaJe w/TR, 1000 for TINI w/TR
    static final int   DTW_SIZE   = 300;    // 300 for linux w/TR, 1000 for SaJe w/TR, 100 for TINI w/TR
    static final int[] dt         = new int[DT_SIZE];   // stores either individual event times or bins of event counts
    static final int[] dtw        = new int[DTW_SIZE];  // stores either individual event times or bins of event counts
    static       int[] dt_b;
    static       int   dt_offset  = 0;    // dt[] offset, ie w/ binning dt[i] correspond to time i+dt_offset
    static     boolean dt_binEvents;      // false -> ind. event times, true -> event cnt bins in dt[]
    static       float nSigma     = 3.5f; // default width of the peak (in stdev)
    static       int   timeoutSec = __TIMEOUT_SEC__;          // timeout, in seconds, for the main timing loop
    static     int MAX_PEAK_WIDTH = 75;   // the timing peak should be found within the first MAX_PEAK_WIDTH bins
    static  double MAX_OK_STDEV   = 4.0;  // maximum value of an ok/good peak stdev
    )
    static      String info;

  public static void main(String[] args)
  {
    int    i;
    int    maxIterations;
    String propertyStr;
    int    ticksPerMilliSecond = 1;
    __DEBUG_TIMING__(
    long   dt_time,  dt_time0;
    int    dt_delta;
    long   dtw_time, dtw_time0;
    int    dtw_delta;
    )
    __DEBUG_MEMORY__(
    //showMemory();
    System.gc(); 
    )
       
    __DEBUG_TIMING__(
    // check to see if the user wants a custom peak width
    propertyStr = System.getProperty("nSigma");
    if (propertyStr != null) {
      nSigma = 0.01f * Integer.parseInt(propertyStr);
      System.out.println("nSigma: " + formatFloat(nSigma));
    }
    )
    
    __DEBUG_TIMING__(
    // check to see if the user wants a custom timeout
    propertyStr = System.getProperty("TO");
    if (propertyStr != null) {
      timeoutSec = Integer.parseInt(propertyStr);
      System.out.println("timeoutSec: " + timeoutSec);
    }
    )
    
    // initialize the timer (actual calls and values supplied by the following macros)
    __TIMER_INIT__;
    ticksPerMilliSecond = __TICKS_PER_MS__;

    Object object;
    C_ORB orb = new C_ORB();
    __DEBUG__( System.out.println("ORB created!");)

    __SECURITY__(
    System.out.println( "\n*** " + orb.cipherEncrypt.toString() + "  keyLen: __SEC_CIPHER_KEY_LEN__");
    // load the 'user' supplied key into the ORB
    byte[] orbKey = {10,11,12,13,14,15,16,17};
    try {
      orb.initCipherKey(orbKey);
    }
    catch (InvalidKeyException e) {
      System.out.println(e);
    }
    )
    __SECURITY_MD__(
    System.out.println("*** Using Message Digest: __SEC_MD_ALG__");
    )
    __SECURITY_MAC__(
    System.out.println("*** Using MAC: __SEC_MAC_ALG__ (__SEC_MAC_ALG_PARAMS__)");
    // load the 'user' supplied MAC key into the ORB
    byte[] macKey = {20,21,22,23,24,25,26,27};
    try {
      orb.initMacKey(macKey);
    }
    catch (InvalidKeyException e) {
      System.out.println(e);
    }
    )

    // Find the server to connect to (via a corbaloc)
    //
    propertyStr = System.getProperty("corbaloc");
    if (propertyStr != null) {
      __DEBUG__( System.out.println("Using 'corbaloc' property: "+propertyStr);)
      object = orb.corbaloc_to_object(propertyStr);
    }
    else {
      __DEBUG__( System.out.println("No 'corbaloc' property, using args[0]: " + args[0]);)
      object = orb.corbaloc_to_object(args[0]);
      //object = orb.string_to_object(args[0]);
    }
    __DEBUG__( System.out.println("corbaloc to object!");)

    // Get the number of iterations
    //
    maxIterations = 1000;  // The default number of iterations
    propertyStr = System.getProperty("cnt");
    if (propertyStr != null) {
      maxIterations = Integer.parseInt(propertyStr);
    }
    System.out.println("Iteration 'cnt' set to " + maxIterations);
    __DEBUG_TIMING__(
    if (maxIterations > DT_SIZE) {
        dt_binEvents = true;      // dt[i] == # of events that took i ms to complete
    } else {
        dt_binEvents = false;     // dt[i] == delta time of event i
    }
    )

    timing.foo fooObj = timing.fooHelper.narrow(object);
    __DEBUG__( System.out.println("Narrow! finished");)

    // configID contains info about the current MQC settings/options
    String configID = "cfg:p" + Config.intPacketSize + ":";
    __SECURITY__(
        configID += orb.cipherEncrypt.toString() + "-__SEC_CIPHER_KEY_LEN__";
    )
    configID += ":__SEC_MAC_ALG__:__SEC_MD_ALG__";

    long ssStartTime = 0;  // steady state start time
    long ssTotalTime = 0;  // steady state total time
    long startTime   = 0;  // start of "real" timing loop time
    long totalTime   = 0;  // total time for the "real" timing loop
    
    //
    // STEADY STATE TIMING LOOP
    //--------------------------
    //
    __DEBUG_MEMORY__(
    //showMemory();
    System.gc();
    )
    
    System.out.println("---Steady State Begin---");
    
    __DEBUG_TIMING__(
    // init variable needed to compute the stdev of time deltas of each call
    for (i = 0;  i < DT_SIZE;  i++) {
      dt[i] = 0;
    }
    for (i = 0;  i < DTW_SIZE;  i++) {
      dtw[i] = 0;
    }
    )
    ssStartTime = __GET_TIME__;
    __DEBUG_TIMING__(
    dt_time0 = __GET_TIME__;
    )
    
    __DEBUG_TIMING__(
    int    dt_min = 999999999;
    int    ss_min;
    int    ss_max;
    int    ss_upperLimit = 999999999;
    )
    int    ss_nLoop = __SS_LOOP_CNT__;
    int    ss_jcnt  = __SS_ITERATIONS__;
    for (int j = 0;  j < ss_nLoop;  j++) {
        __DEBUG_TIMING__(
        dt_sum  = 0;
        dt_sum2 = 0;
        dt_cnt  = 0;
        ss_min  = 999999999;
        ss_max  = 0;
        dt_time0 = __GET_TIME__;
        )
        for (i = 0;  i < ss_jcnt;  i++) {
            fooObj.bar(i);
            __DEBUG_TIMING__(
            dt_time  = __GET_TIME__;
            dt_delta = (int) (dt_time - dt_time0);
            if (dt_delta < ss_min) ss_min = dt_delta;
            if (dt_delta > ss_max) ss_max = dt_delta;
            if (dt_delta < ss_upperLimit) {
                dt_sum  += dt_delta;
                dt_sum2 += (long) dt_delta * (long) dt_delta;
                dt_cnt++;
            }
            dt_time0 = dt_time;
            )
        }
        __DEBUG_TIMING__(
        if (ss_min < dt_min) {
            dt_min = ss_min;
        }
        // semi-fragile code, probably should check that dt_cnt > 1.  -ADM
        dt_avg    = (float) dt_sum / dt_cnt;
        dt_stdev2 = (double)(dt_sum2 - (dt_sum * dt_sum)/dt_cnt) / (dt_cnt-1);
        if (dt_stdev2 >= 0.0 ) {
            dt_stdev = Math.sqrt(dt_stdev2);
        } else {
            System.err.println("Oops! Likely integer overflow -- dt_stdev2= " + 
                            formatDouble(dt_stdev2) + " < 0.0 -- Setting stdev to " + 
                            formatFloat((float)STDEV_ERR) + "!");
            dt_stdev = STDEV_ERR;
        }
        ss_upperLimit = (int) (dt_avg + nSigma * dt_stdev);
        System.out.println(" - " + (j+1)*ss_jcnt + ":   avg: " + formatFloat(dt_avg) + "  stdev: " + formatDouble(dt_stdev) + 
                           "  min/max: " + ss_min + "-" + ss_max + "  upL: " + ss_upperLimit);
        )
    }
    __DEBUG_TIMING__(
    float  dt_ssAvg   = (float) (dt_avg / ticksPerMilliSecond);
    double dt_ssStdev = dt_stdev / ticksPerMilliSecond;
    long   dt_ssCnt   = dt_cnt;
    )
        
    ssTotalTime = (__GET_TIME__ - ssStartTime);
    float ssAvgTime = (float) (ssTotalTime / ticksPerMilliSecond) / (ss_nLoop * ss_jcnt);
    
    __DEBUG_TIMING__(
    System.out.println("---Steady State Info---");
    )
    System.out.println("Overall:  " + formatFloat(ssAvgTime) + "    n/a     " + 
                       (ss_nLoop * ss_jcnt) + "  " + 
                       formatFloat((float)ssTotalTime/(1000*ticksPerMilliSecond)) + "s");
    __DEBUG_TIMING__(
    System.out.println("  ->end:  " + formatFloat(dt_ssAvg) + "  " + 
                       formatDouble(dt_ssStdev) + "  " + dt_ssCnt);
    )
    System.out.println("---Steady State End---" );

    __DEBUG_TIMING__(
    int dd = (int) (5.0 * (dt_ssStdev * ticksPerMilliSecond)) + 10;
    if (dd > (int) (.2 * DT_SIZE)) {
      dd = (int) (.2 * DT_SIZE);
    }
    dt_offset = dt_min - dd;
    if (dt_offset < 0) {
      dt_offset = 0;
    }
    )

    
    //
    // MAIN TIMING LOOP
    //------------------
    //
    int  iterations = 0;
    long stop_time;
    for (int r = 0;  r < REPEAT_CNT;  r++) {
        __DEBUG_TIMING__(
        i = -1;
        dt_delta  = -1;
        dtw_delta = -1;
        try {
        )
            __DEBUG_TIMING__(
            // init variable needed to compute the stdev of time deltas of each call
            for (i = 0;  i < DT_SIZE;  i++) {
                dt[i] = 0;
            }
            for (i = 0;  i < DTW_SIZE;  i++) {
                dtw[i] = 0;
            }
            )
            
            startTime = __GET_TIME__;
            __DEBUG_TIMING__(
            dt_time0  = startTime;
            dtw_time0 = dt_time0 / 1000;
            stop_time = dt_time0 + timeoutSec * 1000 * ticksPerMilliSecond;
            )
            
            int d;
            iterations = maxIterations;
            for (i = 0; i < iterations; i++) {
                fooObj.bar(i);
                __DEBUG_TIMING__(
                dt_time   = __GET_TIME__;
                dt_delta  = (int) (dt_time - dt_time0);
                dt_time0  = dt_time;
                dtw_time  = dt_time / 1000;
                dtw_delta = (int) (dtw_time - dtw_time0);
                dtw_time0 = dtw_time;
                if (dt_binEvents) {
                    d = dt_delta - dt_offset;
                    if (d < DT_SIZE) {
                        dt[d]++;
                    } else {
                        dt[DT_SIZE-1]++;
                    }
                } else {
                    // store raw (unbinned) times into dt[]
                    dt[i]  = dt_delta;
                }
                // wide events are always binned
                if (dtw_delta < DTW_SIZE) {
                    dtw[dtw_delta]++;
                } else {
                    dtw[DTW_SIZE-1]++;
                }
                if (dt_time > stop_time) {
                    iterations = i;
                    break;
                }
                )
            }
            totalTime = (__GET_TIME__ - startTime);
            float avgTime = (float) (totalTime / ticksPerMilliSecond) / iterations;
            float ssDelta = ssAvgTime / avgTime;
        
            __DEBUG_TIMING__(
            dt_stdev = findStdev(false, iterations);  //warning computes dt_cnt, dt_avg, dt_sum, dt_sum2
            dt_avg   /= ticksPerMilliSecond;
            dt_stdev /= ticksPerMilliSecond;
            )
        
            info = "@@@   Loop" + r + ":  " + (int)(totalTime/(1000*ticksPerMilliSecond)) + "  " + formatFloat(avgTime);
            __DEBUG_TIMING__(
            info += "  " + formatDouble(dt_stdev) + "  " + iterations + "  @@  r" + formatFloat(ssDelta);
            )
            info += "  @@  " + configID;
            
            __DEBUG_TIMING__(                  
            dt_stdev  = findStdev(true, iterations);  //warning computes dt_cnt, dt_avg, dt_sum, dt_sum2
            dt_avg   /= ticksPerMilliSecond;
            dt_stdev /= ticksPerMilliSecond;
            float dt_ssDelta = dt_ssAvg / dt_avg;
            )
            
            System.out.println(info);   // print non-GC info
            
            __DEBUG_TIMING__(
            System.out.print("@@@ gcLoop" + r + ":  " + (int)(totalTime/(1000*ticksPerMilliSecond)) + "  " +
                              formatFloat(dt_avg)+ "  " + formatDouble(dt_stdev) + 
                              "  " + dt_cnt + "  @@  r" + formatFloat(dt_ssDelta));
            System.out.println("  @@  " + configID);
            )
            
        __DEBUG_TIMING__(
        }
        catch (java.lang.ArrayIndexOutOfBoundsException e) {
            System.err.println("i: " + i + " dt_offset/dt/dtw: " + dt_offset + 
                               "  " + dt_delta + "  " + dtw_delta);
            dt_cnt   = -1;
            dt_avg   = 999999;
            dt_stdev = STDEV_ERR;
        }
        )
        
        __DEBUG_TIMING__(
        r_cnt[r]   = dt_cnt;
        r_avg[r]   = dt_avg;
        r_stdev[r] = dt_stdev;
        )
    }

    __DEBUG_TIMING__(
    System.out.println("---Statistical Info---");
    System.out.print("Loop s:  " + formatFloat(ssAvgTime));
    System.out.print("  " + formatDouble(dt_ssStdev) + "  " + dt_ssCnt);
    System.out.println();

    dt_cnt   = r_cnt[0];
    dt_avg   = r_avg[0];
    dt_stdev = r_stdev[0];
    for (int r = 0;  r < REPEAT_CNT;  r++) {
        System.out.println("Loop " + r + ":  " + formatFloat(r_avg[r]) +
                           "  " + formatDouble(r_stdev[r]) + "  " + r_cnt[r]);
        if (r_avg[r] < dt_avg) {
            dt_cnt   = r_cnt[r];
            dt_avg   = r_avg[r];
            dt_stdev = r_stdev[r];
        }
    }
    System.out.println("@@@ Summary: " + formatFloat(dt_avg) + "  " +
                       formatDouble(dt_stdev) + "  " +
                       dt_cnt + " @@- " + configID);
    )

    __DEBUG_MEMORY__( 
    showMemory();
    )
    
    __SAJE__(
    // hacks needed to cause the SaJe (Charade) environment to shutdown
    // First, cause the server to shutdown
    fooObj.bar(-1);
    // then it is our turn to shutdown
    shutDown();
    )
  }
    __SAJE__(
    /**
     * shutDown -- Used on the SaJe platform to cause a S/W break
     * point that will timeout a Charade 'run X' command.
     */
    static void shutDown() {
        boolean b = false;
    }
    )

    /**
     * formatFloat -- used to print a float with four decimal digits
     * (This routine is needed because CLDC does not support printing floats).
     *
     * @param f  floating point number to print
     */  
    static String formatFloat(float f) {
        f += 0.00005;
        long fint = (long) f;
        long ffra = (long) (10000 * ((f + 1) - fint));
        StringBuffer ffraStr = new StringBuffer( String.valueOf(ffra));
        ffraStr.setCharAt(0, '.');
        return String.valueOf(fint) + ffraStr;
    }
  
    __DEBUG_TIMING__(
    /**
     * formatDouble -- used to print a double with six decimal digits
     * (This routine is needed because CLDC does not support printing doubles).
     *
     * @param d  double to print
     */  
    static String formatDouble(double d) {
        d += 0.0000005;
        long dint = (long) d;
        long dfra = (long) (1000000 * ((d + 1) - dint));
        StringBuffer dfraStr = new StringBuffer( String.valueOf(dfra));
        dfraStr.setCharAt(0, '.');
        return String.valueOf(dint) + dfraStr;
    }
    )


    __DEBUG_TIMING__(
    /**
     * findStdev -- find the standard deviation of a set of point.
     *
     * @param nonGC   if true, computate the stdev of only the
     *                non-Garbage Collected, if false, then use all events
     * @param iCnt    number of non-binned events in dt[] (see dt_binEvents)
     * @return double the raw standard deviation (ie, non-scaled value).
     */
    static double findStdev(boolean nonGC, int iCnt) {
        //WARNING: global variables dt_cnt, dt_sum, dt_sum2 are all modified
        //within this routine (yes--an ugly hack...)
        
        int    i, iStart, iPeakStart, iStop;
        int    dt_max;
        int    dt_b[];  // bins
        double stdev2, stdev;
        String dt_info;
    
        if (dt_binEvents) {
            // bins already computed, just alias dt[]
            dt_b   = dt;
        } else {
            // need to compute number of bins needed
            dt_offset = 3600000;     // an hour (in ms)
            dt_max    =       0;
            for (i = 0;  i < iCnt;  i++) {
                if (dt[i] < dt_offset) dt_offset = dt[i];
                if (dt[i] > dt_max)    dt_max    = dt[i];
            }
            dt_offset--;
            dt_max++;
            if (dt_max > dt_offset + 2*DT_SIZE ) {
                dt_max = dt_offset + 2*DT_SIZE;
            }
            // create the bins
            dt_b = new int[dt_max - dt_offset + 1];
            // stuff the bins
            for (i = 0;  i < iCnt;  i++) {
                __DEBUG__( if (nonGC) System.err.print(" " + dt[i]); )
                if (dt[i] - dt_offset < 2*DT_SIZE) {
                    dt_b[dt[i] - dt_offset]++;
                } else {
                    dt_b[2*DT_SIZE]++;
                }
            }
            __DEBUG__( if (nonGC) System.err.println(); )
        }
    
        // Initially, use the first MAX_PEAK_WIDTH non-zero channels to compute
        // the average and stdev of the timing peak
        iStart = 0;
        while (dt_b[iStart] == 0  &&  iStart < dt_b.length-1) {
          iStart++;
        }
        iPeakStart = iStart+1;
        while (dt_b[iPeakStart] < 2  &&  iPeakStart < dt_b.length-1) {
          iPeakStart++;
        }

        iStop = dt_b.length;
        if (nonGC) {
            // the last bin contains overflow values--discard it
            iStop--;
        }
        if (iStop > iPeakStart + MAX_PEAK_WIDTH) {
            iStop = iPeakStart + MAX_PEAK_WIDTH;
        }
        if (nonGC) {
            System.err.println("iStart/iPeakStart/iStop: " + (iStart+dt_offset) + 
                       "  " + (iPeakStart+dt_offset) + "  " + (iStop+dt_offset));
        }
        int iStop0;
        long dtb, dti;
        double x, x2;
    
        do {
            iStop0 = iStop;
            // sum up the dt[] bins
            dt_cnt  = 0;
            dt_sum  = 0;
            dt_sum2 = 0;
            for (i = iStart;  i < iStop;  i++) {
                dt_cnt  += dt_b[i];
                dtb = dt_b[i];
                dti = dt_offset + i;
                dt_sum  += dtb * dti;
                dt_sum2 += dtb * dti * dti;
                // check for overflowed values (they will cause dt_sum2 to go negative)
                if (dt_sum2 < 0) {
                    System.err.println("Oops! Likely integer overflow problem.");
                    System.err.println("dt_sum2 < 0: " + dt_sum2 + " at i=" + (dt_offset + i));
                }
            }
      
            // compute the average and its standard dev.
            dt_avg  = (float)   dt_sum  / dt_cnt;
            x  = (double) dt_sum;
            x2 = (double) dt_sum2;
            stdev2  = (x2 - (x * x)/dt_cnt) / (dt_cnt-1);
            if (stdev2 >= 0) {
                stdev = Math.sqrt(stdev2);
            } else {
                System.err.println("Oops! Likely integer overflow -- stdev2= " + 
                                  formatDouble(stdev2) + " < 0.0 -- Setting stdev to " +
                                  formatFloat((float)STDEV_ERR) + "!");
                stdev = STDEV_ERR;
            }
            if (nonGC) {
                iStop = (int) (dt_avg + (nSigma * stdev) + 0.5) - dt_offset;
                if (iStop >= dt_b.length) {
                    iStop = dt_b.length - 1;
                }
                if ((iStop >= iStop0) && (stdev > MAX_OK_STDEV)) {
                    // Compute an alternate upper limit for the first (nonGC)
                    // peak. The value (dt_avg - iStart) is *assumed* to be the
                    // bottom (lower-time) tail of the nonGC peak. This means 
                    // that (dt_avg - iStart) should be another approximation
                    // to the value of 3.5 * nSigma (of the nonGC peak). If this
                    // value gives a smaller iStop use it.
                    int iStop2 = (int) ((dt_avg - dt_offset) + ((dt_avg - dt_offset) - iPeakStart));
                    System.err.println("avg,iStop,iStop2: " + formatFloat((float)dt_avg) + "  " +
                                       (dt_offset+iStop) + "  " + (dt_offset+iStop2));
                    if (iStop2 < iStop) {
                        iStop = iStop2;
                    }
                }
                if (iStop < iStart + 1) {
                    iStop = iStart + 1;
                }
                System.err.println("x-y,avg,stdev,iStop: " + 
                    (dt_offset+iStart) + "-" + (dt_offset+iStop0) + "  " + 
                    formatDouble(dt_avg) + "  " + formatDouble(stdev) + 
                    "  " + (dt_offset+iStop));
            }
        } while (iStop < iStop0);
        
        // Pretty-Print the data
        if (nonGC) {
            System.out.println("--- " + (dt_offset+iStart) + "  " + 
                               (dt_offset+(iStop0-1)) + " ---non-GC---");
            printArray(dt_b, dt_offset);
            System.out.println("---Wide bins---");
            printArray(dtw, 0);
        }
    
        if (! dt_binEvents) {
            dt_b = null;    // release the dt_b array--it is no longer needed
        }
        
        return stdev;
    }

    /**
    * printArray -- a pretty printer for the dt/dtw arrays
    *
    * @param a     an array to print
    * @param offset  index offset of the array
    */
    static void printArray(int a[], int offset)
    {
        int     lastZero = 0;
        boolean inZeros  = true;
        System.out.println( offset + "  " + a[0]);
        for (int i = 1;  i < a.length;  i++) {
            if (a[i] > 0) {
                if (inZeros) {
                    inZeros = false;
                    if (i-1 > lastZero) {
                        System.out.println((offset+i-1) + "  " + a[i-1]);
                    }
                }
                System.out.println((offset+i) + "  " + a[i]);
            } else {
                if ( inZeros ) {
                    // do nothing
                } else {
                    inZeros  = true;
                    lastZero = i;
                    System.out.println((offset+i)+ "  " + a[i]);
                }
            }
        }
    }
    )

    __DEBUG_MEMORY__(
    static Runtime runtime = Runtime.getRuntime();
    
    /**
     * showMemory -- print our the free/used memory
     */
    static void showMemory() {
        long tMem, fMem, uMem, fMem1, uMemMax, fMemMax;
    
        tMem    = runtime.totalMemory();
        fMem    = runtime.freeMemory();
        fMemMax = fMem;
        if (tMem == fMem) {
            // TINI hack, tMem and fMem are reported as equal
            tMem = 334240;  // MAGIC NUMBER appears to be the proper value
        }
        uMem    = tMem - fMem;
        __DEBUG__( System.out.println("c.Memory (t/f/u): \t"+tMem+" \t"+fMem+" \t"+uMem);)
    
        System.gc();
        
        for (int i = 0;  i < 100;  i++) {
            fMem1 = runtime.freeMemory();
            __DEBUG__( System.out.println("c.fMem: " + fMem1););
      
            if (fMem1 > fMemMax ) {
                fMemMax = fMem1;
            } else {
                if (i > 1) break;
            }
            try {
                Thread.sleep(500);
            } catch (InterruptedException ie) {}
        }
        uMemMax = tMem - fMemMax;
    
        System.out.println("@@@ c.Memory (t/f/u/du): \t"
                           + tMem + " \t" + fMemMax + " \t" + uMemMax + " \t" + (uMemMax - uMem));
    }
    )

}
