/*
 * Copyright (c) 2003 David E. Bakken, his research students, and Washington State University.
 * Please see the file LICENSE.pdf for more details on terms of use.
 */

package mqc;

import mqc.holders.*;

import mqc.protocols.*;

import java.io.*;
import java.util.*;

/**
* Base class that all stubs inherit from. Contains
* functions to get and set the delegate.
*
* @author Olav Haugan
*
* 14.10.01
*/
public class ObjImpl
{
   private Delegate delegate;

   /**
   * Sets the delegate.
   *
   * @param Delegate The delegate object to be assigned.
   */
   public void set_delegate(Delegate delegate)
   {
      this.delegate = delegate;
   }

   /**
   * Gets the delegate.
   *
   * @return Delegate The delegate object
   */
   public Delegate get_delegate()
   {
      return delegate;
   }

   /**
   * Creates an request message using CDRStream.
   *
   * @param String The name of the operation requested.
   * @param Boolean true if oneway call, false otherwise.
   * @return CDROutputStream The output stream.
   */
   public CDROutputStream _request(String operation,
                                   boolean twoWay)
   {
      return delegate._request(operation, twoWay, 0);
   }

   /**
   * Creates an request message using a byte buffer.
   *
   * @param String The name of the operation requested.
   * @param int Size of buffer needed.
   * @param Boolean true if oneway call, false otherwise.
   * @param IntHolder The buffer pointer
   * @return byte[] The buffer
   */
   public byte[] _request(String operation,
                          int sizeOfBuffer,
                          boolean twoWay,
                          IntHolder ptr)
   {
      return delegate._request(operation, sizeOfBuffer, twoWay, ptr);
   }

   /**
   * Send the request to the server and awaits an answer
   * using CDRStreams.
   *
   * @param CDROutputStream The output stream.
   * @return CDRInputStream The input stream.
   */
   public CDRInputStream _invoke(CDROutputStream output)
   {
      return delegate._invoke(output);
   }

   /**
   * Send the request to the server and awaits an answer
   * using byte buffers.
   *
   * @param byte[] The output stream.
   * @param IntHolder The buffer pointer
   * @return byte[] The input stream.
   */
   public byte[] _invoke(byte[] buffer, IntHolder ptr, boolean twoWay)
   {
      return delegate._invoke(buffer, ptr, twoWay);
   }
}
