/*
 * Copyright (c) 2003 David E. Bakken, his research students, and Washington State University.
 * Please see the file LICENSE.pdf for more details on terms of use.
 */

package mqc.protocols;

/**
 * Hello.
 *
 * struct GIOPLiteHeader {
 *    unsigned long    msg_size;
 *    octet            msg_type;
 * };
 *
 */
public class GIOPLiteHeader implements Marshalable
{
   /* GIOPLite does not have the magic field in the header! */
   public static final int SIZE_OF_HDR = 8;

   protected int m_size;
   protected byte m_type;

   public GIOPLiteHeader( byte type )
   {
      this( 0, type );
   }

   public GIOPLiteHeader( int size, byte type )
   {
      m_size = size;
      m_type = type;
   }

   public int getSize()
   {
      return m_size;
   }

   public void setSize( int size )
   {
      m_size = size;
   }

   public byte getType()
   {
      return m_type;
   }

   public void setType( byte type )
   {
      m_type = type;
   }

   public void marshal( CDROutputStream out)
   {
      // unsigned long msg_size;
      out.writeUnsignedLong( m_size );

      // octet msg_type
      out.writeOctet( m_type );
   }

   public void deMarshal( CDRInputStream in )
   {
      // unsigned long msg_size;
      m_size = in.readUnsignedLong();
      
      // octet msg_type
      m_type = in.readOctet();
   }

   public static int getMsgSize( byte[] buf, boolean bigEndian )
   {
      return CDRInputStream.readInt( buf, 0, bigEndian );
   }

   public static void setMsgSize( byte[] buf, int value, boolean bigEndian )
   {
      CDROutputStream.writeInt( buf, 0, value, bigEndian );
   }

}
