/*
 * Copyright (c) 2003 David E. Bakken, his research students, and Washington State University.
 * Please see the file LICENSE.pdf for more details on terms of use.
 */


package mqc.udp;

import mqc.*;
import java.io.*;
import java.net.*;
import java.util.*;
import java.util.Date;
import java.util.LinkedList;
import javax.swing.Timer;
import java.lang.*;
import java.awt.event.*;

public class UDPGoBackNSocket extends UDPSocket
{
   private final int d_list_length = 10;
   private final int d_time_out = 2000;


   //Table of N
   private java.util.List d_list;

   //Receiver Thread
   private UDPReceiver receiver;

   //Packet number
   private short packet_num;

   //Recieved packet info
   public boolean newPacket;
   public byte receivedMsg[];
   public int msgSize;

   //Class containing info
   class PacketData
   {
      public DatagramPacket packet;
      public short seq_num;
      public long time_stamp;
   }


   public UDPGoBackNSocket(DatagramSocket socketIn, String server, int server_port)
   {
      super(socketIn, server, server_port);

      //Allocate memory
      //receivedMsg = new byte[1024];
      receivedMsg = null;

      //Set the new msg as false
      newPacket = false;

      //Set the packet num to zero
      packet_num = 0;

      //Create the list
      d_list = Collections.synchronizedList(new LinkedList());

      //Create the receive thread
      receiver = new UDPReceiver( this );

      //Start the receiver thread
      receiver.start();

      ActionListener taskPerformer = new ActionListener()
                                     {
                                        public void actionPerformed(ActionEvent e )
                                        {
                                           PacketData data;
                                           long cur_time;

                                           //Get the last element
                                           if ( d_list.size() > 0 )
                                              data = (PacketData) d_list.get(d_list.size()-1);
                                           else
                                              return;

                                           //Get the current time
                                           cur_time = System.currentTimeMillis();

                                           if ( cur_time - data.time_stamp > d_time_out )
                                           {
                                              timeOut();
                                           }
                                        }
                                     };

      Timer timer = new Timer(20, taskPerformer);

      timer.start();

   }

   public void timeOut()
   {
      PacketData data;

      //Traverse through the list
      for (int i = d_list.size() - 1; i >= 0; --i)
      {
         //Get the data
         data = (PacketData) d_list.get(i);

         //Send the the packet
         sendPacket(data.packet);

         //Reset the time
         data.time_stamp = System.currentTimeMillis();
      }
   }

   public synchronized void removeFrom( short num )
   {
      PacketData data;
      int size;
      int lastIndex;
      int i = 0;

      //Search in the list data with this seq_num
      data = search(num);

      //If Search doesnot fail
      if ( data != null )
      {
         lastIndex = d_list.indexOf(data);
      }
      else
         return;

      //Set size
      size = d_list.size();

      //Set i as last index
      i = lastIndex;

      //Remove everything below this data
      for ( ; i < size; ++i )
      {
         d_list.remove(lastIndex);
      }
   }

   public synchronized void addPacketData( DatagramPacket packet, short num )
   {
      PacketData packetData = new PacketData();

      packetData.packet = packet;
      packetData.seq_num = num;
      packetData.time_stamp = System.currentTimeMillis();

      //Add in the front of the list
      d_list.add(0, packetData);
   }

   public synchronized PacketData search(short number)
   {
      PacketData packetData;

      //Traverse through the list
      for (int i = 0; i < d_list.size(); ++i)
      {
         //Get the data
         packetData = (PacketData)d_list.get(i);

         //If seq number of the packet is equal to the given number
         if ( packetData.seq_num == number )
         {
            return packetData;
         }
      }

      return null;
   }

   public void send(byte [] data, int size)
   {

      DatagramPacket packetToSend;

      //Increment the packet
      ++packet_num;

      //Prepare the packet
      packetToSend = preparePacket(data, size, packet_num);

      //Check the size of the list. If more then return
      if ( d_list.size() > d_list_length )
      {
         System.out.println("Buffer full!!!");
         try
         {
            while ( d_list.size() > d_list_length )
            {

               Thread.sleep(20);
            }

         }
         catch ( Exception e )
         {
            System.out.println(e.getMessage());
         }
      }

      //Add the packet to the list
      addPacketData( packetToSend, packet_num);

      //Send the packet
      sendPacket( packetToSend);
   }

   public synchronized int receive(byte[] data)
   {
      receivedMsg = data;

      //receiver.receiverNotify();

      //Wait till u get the next packet
      try
      {
         wait();
      }
      catch( Exception e)
      {
         System.out.println("Error");
         System.out.println(e.getMessage());
      }

      receivedMsg = null;
      newPacket = false;

      //return the msg size
      return msgSize;
   }

   public synchronized void sendPacket(DatagramPacket packetIn)
   {
      //int random_num;

      //random_num = (int)(Math.random()*10000.0);

      //Send the packet
      try
      {
         //if ( random_num > 9 )
         d_socket.send(packetIn);

      }
      catch(Exception e )
      {
         System.out.println(e.getMessage());
      }
   }

   public synchronized DatagramPacket preparePacket(byte[] data, int size, short number)
   {
      //byte msg_data[] = new byte[1024];
      byte msg_data[] = new byte[Config.intPacketSize * 2];
      DatagramPacket packet;

      //Convert short to byte
      msg_data[0] = (byte) (number >> 8);
      msg_data[1] = (byte) number;

      //Add to the array of byte
      System.arraycopy(data, 0, msg_data, 2, size );

      //Create the packet
      packet = new DatagramPacket( msg_data, size+(short)2, server_address, server_port);

      return packet;
   }

   public synchronized void socketNotify()
   {
      notify();
   }
}

class UDPReceiver extends java.lang.Thread
{

   private short d_last_msg_num;

   private boolean isConnected;

   private UDPGoBackNSocket socket;

   private int msg_size;

   private short msg_seq_num;

   private byte data[];

   public UDPReceiver(  UDPGoBackNSocket socket_in)
   {
      d_last_msg_num = 0;

      isConnected = true;

      //data = new byte[1024];
      data = new byte[Config.intPacketSize * 2];

      socket = socket_in;
   }

   public void extractPacketInfo(DatagramPacket packet_in, byte[] data )
   {
      byte seq_byte[] = new byte[2];
      byte [] receive_data;


      //Get the data and its size
      receive_data = packet_in.getData();
      msg_size = packet_in.getLength();

      //Read the first few bytes of the data to get the sequence no.
      seq_byte[0] = receive_data[0];
      seq_byte[1] = receive_data[1];

      //Get the sequence num
      msg_seq_num = (short) (((seq_byte[0] & 0xFF) << 8) | (seq_byte[1] & 0xFF));

      //System.out.println(msg_seq_num);

      //Update the size
      msg_size = msg_size - 2;

      System.arraycopy(receive_data, 2, data, 0, msg_size );
   }

   public void run()
   {

      DatagramPacket packet;
      byte data[] = new byte[Config.intPacketSize];
      int ret_value;
      DatagramPacket ack;

      //While it is connected
      while ( isConnected == true )
      {
         //Create a packet to receive
         packet = new DatagramPacket( data, data.length );

         //Receiver Packet
         try
         {
            socket.d_socket.receive(packet);
            //System.out.println("Received");

         }
         catch (Exception e )
         {
            System.out.println(e.getMessage());
         }

         //Extract information from the packet
         extractPacketInfo(packet, data);

         if ( msg_seq_num < 0 )
         {
            //Packet is an Ack
            handleAck((short)-msg_seq_num);
         }
         else
         {
            if ( msg_seq_num == d_last_msg_num + 1 /*tbd: check for 2^15*/ )
            {

               //Packet is a msg
               handleMsg(data, msg_size, msg_seq_num);

               //Set the last msg num
               d_last_msg_num = msg_seq_num;
            }
            else
            {
               //Send the ack for the last message in case this is a duplicate message
               //prepare an ack
               ack = socket.preparePacket( data, 0, (short)-d_last_msg_num );


               //Send the ack
               socket.sendPacket(ack);

            }

         }//End of if statement
      }//End of While loop
   }//End of function


   public synchronized void handleAck(short number )
   {
      //Remove the packets with seq num and those following
      socket.removeFrom(number);
   }

   public synchronized void handleMsg(byte [] msg, int size, short number )
   {
      DatagramPacket ack;
      byte data[] = new byte[1];

      if ( socket.receivedMsg == null )
      {
         //System.out.println("msg null");
         while ( socket.receivedMsg == null );
         try
         {
            //wait();
         }
         catch (Exception e )
         {
            System.out.println(e.getMessage());
         }
      }


      //Copy the array
      System.arraycopy(msg, 0, socket.receivedMsg, 0, size );

      //Set the msg size
      socket.msgSize = size;

      //Set the new msg to be true
      socket.newPacket = true;

      socket.socketNotify();

      //Send an ack
      //Prepare an ack
      ack = socket.preparePacket( data, 0, (short)-number );

      //Send the ack
      socket.sendPacket(ack);
   }

   public synchronized void receiverNotify()
   {
      //System.out.println("Notify");
      //notify();
   }
}
