/*
 * Copyright (c) 2003 David E. Bakken, his research students, and Washington State University.
 * Please see the file LICENSE.pdf for more details on terms of use.
 */


package mqc.udp;

import mqc.*;
import java.io.*;
import java.net.*;
import java.util.*;
import java.util.Date;
import java.util.LinkedList;
import javax.swing.Timer;
import java.lang.*;
import java.awt.event.*;

/**
 * UDPStopAndWait: implements a basic reliable UDP Socket using Stop and Wait
 */
public class UDPStopAndWaitSocket extends UDPSocket
{
   private final int waittime = 1000; //Miliseconds


   /**
    * Constrcutor that takes in the udp socket to be used for communication,
    * server name & server port
    */
   public UDPStopAndWaitSocket(DatagramSocket socketIn, String server, int server_port)
   {
      super(socketIn, server, server_port);

   }

   /**
    * sends the data inform of udp packet
    */
   public void send(byte [] data, int size)
   {
      boolean ackReceived = false;
      byte [] junk = new byte[Config.intPacketSize * 2];
      DatagramPacket packet = new DatagramPacket(junk, junk.length);

      //int random_num;

      //Set a finite wait time for resending the packet in necessary
      try
      {
         d_socket.setSoTimeout(waittime);
      }
      catch ( Exception e )
      {
         System.out.println(e.getMessage());
      }
      while ( ackReceived != true )
      {


         //random_num = (int)(Math.random()*10000.0);

         try
         {
            //Send a udp packet
            //if ( random_num > 999 )
            d_socket.send( new DatagramPacket(data, size, server_address, server_port ));

            //Wait for a reply
            d_socket.receive(packet);

            ackReceived = true;
         }
         catch ( SocketTimeoutException e )
         {
            //System.out.println("Timeout.. resending the packet");
            ackReceived = false;
            continue;
         }
         catch (Exception e )
         {
            System.out.println("Error sending msg");
            ackReceived = true;
         }
      }
   }

   /**
    * receives a udp packet and passes on the information in data
    */ 
   public int receive(byte[] data)
   {
      boolean packetReceived = false;
      byte [] junk = new byte[Config.intPacketSize * 2];

      //Create a udp packet to receive in the information
      DatagramPacket packet = new DatagramPacket(data, data.length);

      //Set a infinite wait time so that the socket doesnot time out at receive
      try
      {
         d_socket.setSoTimeout(0);
      }
      catch ( Exception e )
      {
         System.out.println(e.getMessage());
      }
      try
      {
         //Receive the packet
         d_socket.receive(packet);

         //Send junk as an ack
         d_socket.send( new DatagramPacket(junk, 1, server_address, server_port ));
      }
      catch(Exception e)
      {
         System.out.println("Error receiving data from socket");
         return 0;
      }

      return packet.getLength();
   }
}
